/*++

// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (c) Microsoft Corporation. All rights reserved

Module Name:

    WpdObjectProperties.cpp

Abstract:

    This module contains the implementation for the 
    WpdObjectProperties class
--*/

#include "internal.h"
#include <math.h>
#include "WpdObjectProperties.h"
#include "SensorDriver.h"
#include "SensorDevice.h"

#include "WpdObjectProperties.tmh"

extern WCHAR g_wszDeviceName[MAX_PATH];

const PROPERTYKEY g_SupportedCommonProperties[] =
{
    WPD_OBJECT_ID,
    WPD_OBJECT_PERSISTENT_UNIQUE_ID,
    WPD_OBJECT_PARENT_ID,
    WPD_OBJECT_NAME,
    WPD_OBJECT_FORMAT,
    WPD_OBJECT_CONTENT_TYPE,
    WPD_OBJECT_CAN_DELETE
};

const PROPERTYKEY g_SupportedDeviceProperties[] =
{
    WPD_DEVICE_FIRMWARE_VERSION,
    WPD_DEVICE_POWER_LEVEL,
    WPD_DEVICE_POWER_SOURCE,
    WPD_DEVICE_PROTOCOL,
    WPD_DEVICE_MODEL,
    WPD_DEVICE_SERIAL_NUMBER,
    WPD_DEVICE_SUPPORTS_NON_CONSUMABLE,
    WPD_DEVICE_MANUFACTURER,
    WPD_DEVICE_FRIENDLY_NAME,
    WPD_DEVICE_TYPE,
    WPD_FUNCTIONAL_OBJECT_CATEGORY
};


/**
 *  This method is called to copy PROPERTYKEYs from one collection to the other.
 *
 *  The parameters sent to us are:
 *  pSourceKeys - The source collection
 *  pTargetKeys - The target collection
 *
 *  The driver should:
 *  Copy PROPERTYKEYs from the source collection to the target.
 */
HRESULT CopyPropertyKeys(
    IPortableDeviceKeyCollection *pSourceKeys,
    IPortableDeviceKeyCollection *pTargetKeys)
{
    HRESULT hr = S_OK;
    DWORD cKeys = 0;
    PROPERTYKEY key = {0};

    if (NULL != pSourceKeys && NULL != pTargetKeys)
    {
        hr = pSourceKeys->GetCount(&cKeys);
        
        if (SUCCEEDED(hr))
        {
            for (DWORD dwIndex = 0; dwIndex < cKeys; ++dwIndex)
            {
                hr = pSourceKeys->GetAt(dwIndex, &key);
                
                if (SUCCEEDED(hr))
                {
                    hr = pTargetKeys->Add(key);
                }
            }
        }
    }
    else
    {
        hr = E_POINTER;
    }

    return hr;
}


HRESULT FindSensorTypeFromObjectID(__in LPWSTR pwszObjectID, 
                                   __out SensorType* pSensorType)
{
    HRESULT hr = S_OK;
    CAtlStringW strObjectID = pwszObjectID;

    if (strObjectID.CompareNoCase(SENSOR_ACCELEROMETER_ID) == 0)
    {
        *pSensorType = Accelerometer;
    }
    else if (strObjectID.CompareNoCase(SENSOR_AMBIENT_LIGHT_ID) == 0)
    {
        *pSensorType = AmbientLight;
    }
    else if (strObjectID.CompareNoCase(SENSOR_SWITCH_ARRAY_LEFT_ID) == 0)
    {
        *pSensorType = SwitchArrayLeft;
    }
    else if (strObjectID.CompareNoCase(SENSOR_SWITCH_ARRAY_RIGHT_ID) == 0)
    {
        *pSensorType = SwitchArrayRight;
    }
    else
    {
        hr = HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED);
    }

    return hr;
}

/**
 *  This method is called to determine if the given PROPERTYKEY is a settable
 *  property of the SENSOR object.
 *
 *  The parameters sent to us are:
 *  Key - A PROPERTYKEY
 *
 *  The driver should:
 *  Return TRUE if the PROPERTYKEY is supported; FALSE otherwise.
 */
BOOL IsSettablePropertyKey(PROPERTYKEY Key)
{
    HRESULT hr = S_OK;

    CComPtr<IPortableDeviceKeyCollection> pSensorProperties;
    
    BOOL fResult = FALSE;

    // Check the settable properties
    // All the sensors have the same settable properties (Use Accelerometer for now)
    hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(Accelerometer))->GetSettableProperties(&pSensorProperties);
        
    if (SUCCEEDED(hr))
    {
        fResult = FindPropertyKey(Key, pSensorProperties);
    }

    return fResult;
}

/**
 *  This method is called to determine if the given PROPERTYKEY can be
 *  found in the given collection
 *
 *  The parameters sent to us are:
 *  Key - A PROPERTYKEY
 *  pKeys - The collection to search in.
 *
 *  The driver should:
 *  Return TRUE if the PROPERTYKEY is found; FALSE otherwise.
 */
BOOL FindPropertyKey(PROPERTYKEY Key, IPortableDeviceKeyCollection *pKeys)
{
    HRESULT hr = S_OK;

    DWORD cKeys = 0;
    PROPERTYKEY tempKey;
    BOOL fResult = FALSE;

    if (NULL != pKeys)
    {
        hr = pKeys->GetCount(&cKeys);
        
        if (SUCCEEDED(hr))
        {
            for (DWORD dwIndex = 0; dwIndex < cKeys; ++dwIndex)
            {
                hr = pKeys->GetAt(dwIndex, &tempKey);
                
                if (SUCCEEDED(hr))
                {
                    if (IsEqualPropertyKey(tempKey, Key))
                    {
                        fResult = TRUE;
                        break;
                    }
                }
            }
        }
    }

    return fResult;
}




/**
 *  This method is called to populate common PROPERTYKEYs found on ALL objects.
 *
 *  The parameters sent to us are:
 *  pKeys - An IPortableDeviceKeyCollection to be populated with PROPERTYKEYs
 *
 *  The driver should:
 *  Add PROPERTYKEYs pertaining to the ALL objects.
 */
VOID AddCommonPropertyKeys(
    IPortableDeviceKeyCollection* pKeys)
{
    if (pKeys != NULL)
    {
        for (DWORD dwIndex = 0; dwIndex < ARRAYSIZE(g_SupportedCommonProperties); dwIndex++)
        {
            pKeys->Add(g_SupportedCommonProperties[dwIndex] );
        }
    }
}

/**
 *  This method is called to populate common PROPERTYKEYs found on the DEVICE object.
 *
 *  The parameters sent to us are:
 *  pKeys - An IPortableDeviceKeyCollection to be populated with PROPERTYKEYs
 *
 *  The driver should:
 *  Add PROPERTYKEYs pertaining to the DEVICE object.
 */
VOID AddDevicePropertyKeys(
    IPortableDeviceKeyCollection* pKeys)
{
    if (pKeys != NULL)
    {
        for (DWORD dwIndex = 0; dwIndex < ARRAYSIZE(g_SupportedDeviceProperties); dwIndex++)
        {
            pKeys->Add(g_SupportedDeviceProperties[dwIndex] );
        }
    }
}

/**
 *  This method is called to populate common PROPERTYKEYs found on ambient light sensor objects.
 *
 *  The parameters sent to us are:
 *  pKeys - An IPortableDeviceKeyCollection to be populated with PROPERTYKEYs
 *
 *  The driver should:
 *  Add PROPERTYKEYs pertaining to the ambient light sensor objects.
 */
HRESULT AddSensorPropertyKeys(
    SensorType sensType,
    IPortableDeviceKeyCollection* pKeys)
{
    HRESULT hr = S_OK;

    CComPtr<IPortableDeviceKeyCollection> pSensorProperties;
    
    if(SUCCEEDED(hr))
    {
        if ((NULL != pKeys))
        {
            // Add the supported properties
            hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(sensType))->GetSupportedProperties(&pSensorProperties);
            
            if (SUCCEEDED(hr))
            {
                hr = CopyPropertyKeys(pSensorProperties, pKeys);
            }
        }
    }

    return hr;
}

/**
 *  This method is called to populate common PROPERTYKEYs found on ambient light sensor objects.
 *
 *  The parameters sent to us are:
 *  pKeys - An IPortableDeviceKeyCollection to be populated with PROPERTYKEYs
 *
 *  The driver should:
 *  Add PROPERTYKEYs pertaining to the ambient light sensor objects.
 */
HRESULT AddSensorDataFieldKeys(
       SensorType sensType,
       IPortableDeviceKeyCollection* pKeys)
{    
    HRESULT hr = S_OK;

    CComPtr<IPortableDeviceKeyCollection> pSensorDataFields;

    if (NULL != pKeys)
    {
        // Add the supported properties
        hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(sensType))->GetSupportedDataFields(&pSensorDataFields);
        
        if (SUCCEEDED(hr))
        {
            hr = CopyPropertyKeys(pSensorDataFields, pKeys);
        }
    }

    return hr;
   
}

/**
 *  This method is called to populate supported PROPERTYKEYs found on objects.
 *
 *  The parameters sent to us are:
 *  wszObjectID - the object whose supported property keys are being requested
 *  pKeys - An IPortableDeviceKeyCollection to be populated with supported PROPERTYKEYs
 *
 *  The driver should:
 *  Add PROPERTYKEYs pertaining to the specified object.
 */
HRESULT AddSupportedPropertyKeys(
    __in LPWSTR                        wszObjectID,
    __out IPortableDeviceKeyCollection*  pKeys)
{
    HRESULT hr = S_OK;
    CAtlStringW strObjectID = wszObjectID;
  
    // Add Common PROPERTYKEYs for ALL WPD objects
    AddCommonPropertyKeys(pKeys);

    if (strObjectID.CompareNoCase(WPD_DEVICE_OBJECT_ID) == 0)
    {
        // Add the PROPERTYKEYs for the 'DEVICE' object
        AddDevicePropertyKeys(pKeys);
    }

    // Add the PROPERTYKEYs for the 'SENSOR' object
    SensorType sType;
    hr = FindSensorTypeFromObjectID(wszObjectID, &sType);
    if(SUCCEEDED(hr))
    {
        hr = AddSensorPropertyKeys(sType, pKeys);
    }

    return hr;
}

/**
 *  This method is called to populate supported PROPERTYKEYs found on data objects.
 *
 *  The parameters sent to us are:
 *  wszObjectID - the object whose supported property keys are being requested
 *  pKeys - An IPortableDeviceKeyCollection to be populated with supported PROPERTYKEYs
 *
 *  The driver should:
 *  Add PROPERTYKEYs pertaining to the specified object.
 */
HRESULT AddSupportedDataFieldKeys(
    __in  LPWSTR                        wszObjectID,
    __out IPortableDeviceKeyCollection*  pKeys)
{
    HRESULT     hr          = S_OK;
    CAtlStringW strObjectID = wszObjectID;

    // Add the DATAFIELD keys for the 'SENSOR' object
    SensorType sType;
    hr = FindSensorTypeFromObjectID(wszObjectID, &sType);
    if(SUCCEEDED(hr))
    {
        hr = AddSensorDataFieldKeys(sType, pKeys);
    }
    
    return hr;
}

WpdObjectProperties::WpdObjectProperties()
{

}

WpdObjectProperties::~WpdObjectProperties()
{

}

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_GET_SUPPORTED
 *  command.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose supported properties have
 *      been requested.
 *
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_FILTER: the filter to use when returning supported properties.
 *      Since this parameter is optional, it may not exist.
 *      ! This driver currently ignores the filter parameter. !
 *
 *  The driver should:
 *  - Return supported property keys for the specified object in WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_KEYS
 */
HRESULT WpdObjectProperties::OnGetSupportedProperties(
        __in LPWSTR  wszObjectID,
        IPortableDeviceKeyCollection** ppKeys )
{
    HRESULT hr          = S_OK;

    // CoCreate a collection to store the supported property keys.
    hr = CoCreateInstance(CLSID_PortableDeviceKeyCollection,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceKeyCollection,
                              (VOID**) ppKeys);

    // Add supported property keys for the specified object to the collection
    if (hr == S_OK)
    {
        hr = AddSupportedPropertyKeys(wszObjectID, *ppKeys);
    }

    return hr;
}

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_GET_SUPPORTED
 *  command for the data child object of the wszObjectID.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose supported properties have
 *      been requested.
 *
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_FILTER: the filter to use when returning supported properties.
 *      Since this parameter is optional, it may not exist.
 *      ! This driver currently ignores the filter parameter. !
 *
 *  The driver should:
 *  - Return supported property keys for the specified object in WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_KEYS
 */
HRESULT WpdObjectProperties::OnGetSupportedDataFields(
        __in LPWSTR  wszObjectID,
        IPortableDeviceKeyCollection** ppKeys )
{
    HRESULT hr          = S_OK;

    // CoCreate a collection to store the supported property keys.
    hr = CoCreateInstance(CLSID_PortableDeviceKeyCollection,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceKeyCollection,
                              (VOID**) ppKeys);

    // Add supported property keys for the specified object to the collection
    if (hr == S_OK)
    {
        hr = AddSupportedDataFieldKeys(wszObjectID, *ppKeys);
    }

    return hr;
}

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_GET
 *  command.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose property values have been requested.
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_KEYS: a collection of property keys, identifying which
 *      specific property values we are requested to return.
 *
 *  The driver should:
 *  - Return all requested property values in WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_VALUES.  If any property read failed, the corresponding value should be
 *      set to type VT_ERROR with the 'scode' member holding the HRESULT reason for the failure.
 *  - S_OK should be returned if all properties were read successfully.
 *  - S_FALSE should be returned if any property read failed.
 *  - Any error return indicates that the driver did not fill in any results, and the caller will
 *      not attempt to unpack any property values.
 */
HRESULT WpdObjectProperties::OnGetPropertyValues(
        __in LPWSTR  wszObjectID,
        IPortableDeviceKeyCollection* pKeys,
        IPortableDeviceValues** ppValues )
{
    HRESULT hr          = S_OK;

    // CoCreate a collection to store the property values.
    hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**) ppValues);

    // Read the specified properties on the specified object and add the property values to the collection.
    if (hr == S_OK)
    {
        hr = GetPropertyValuesForObject(wszObjectID, pKeys, *ppValues);
    }

    return hr;
}

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_GET
 *  command for the data child object of wszObjectID.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose property values have been requested.
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_KEYS: a collection of property keys, identifying which
 *      specific property values we are requested to return.
 *
 *  The driver should:
 *  - Return all requested property values in WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_VALUES.  If any property read failed, the corresponding value should be
 *      set to type VT_ERROR with the 'scode' member holding the HRESULT reason for the failure.
 *  - S_OK should be returned if all properties were read successfully.
 *  - S_FALSE should be returned if any property read failed.
 *  - Any error return indicates that the driver did not fill in any results, and the caller will
 *      not attempt to unpack any property values.
 */
HRESULT WpdObjectProperties::OnGetDataFields(
        __in LPWSTR  wszObjectID,
        IPortableDeviceKeyCollection* pKeys,
        IPortableDeviceValues** ppValues )
{
    HRESULT hr          = S_OK;

    // CoCreate a collection to store the property values.
    hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**) ppValues);

    // Read the specified properties on the specified object and add the property values to the collection.
    if (hr == S_OK)
    {
        hr = GetDataValuesForObject(wszObjectID, pKeys, *ppValues);
    }

    return hr;
}

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_GET_ALL
 *  command.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose property values have been requested.
 *
 *  The driver should:
 *  - Return all property values in WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_VALUES.  If any property read failed, the corresponding value should be
 *      set to type VT_ERROR with the 'scode' member holding the HRESULT reason for the failure.
 *  - S_OK should be returned if all properties were read successfully.
 *  - S_FALSE should be returned if any property read failed.
 *  - Any error return indicates that the driver did not fill in any results, and the caller will
 *      not attempt to unpack any property values.
 */
/*
HRESULT WpdObjectProperties::OnGetAllPropertyValues(
    IPortableDeviceValues*  pParams,
    IPortableDeviceValues*  pResults)
{
    HRESULT hr          = S_OK;
    LPWSTR  wszObjectID = NULL;
    CComPtr<IPortableDeviceValues>        pValues;
    CComPtr<IPortableDeviceKeyCollection> pKeys;

    // First get ALL parameters for this command.  If we cannot get ALL parameters
    // then E_INVALIDARG should be returned and no further processing should occur.

    // Get the object identifier whose property values have been requested
    if (hr == S_OK)
    {
        hr = pParams->GetStringValue(WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID, &wszObjectID);
    }

    // CoCreate a collection to store the property values.
    if (hr == S_OK)
    {
        hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**) &pValues);
    }

    // CoCreate a collection to store the property keys we are going to use
    // to request the property values of.
    if (hr == S_OK)
    {
        hr = CoCreateInstance(CLSID_PortableDeviceKeyCollection,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceKeyCollection,
                              (VOID**) &pKeys);
    }

    // First we make a request for ALL supported property keys for the specified object.
    // Next, we delegate to our helper function GetPropertyValuesForObject( ) passing
    // the entire property key collection.  This will reuse existing implementation
    // in our driver to perform the GetAllPropertyValues operation.
    if (hr == S_OK)
    {
        hr = AddSupportedPropertyKeys(wszObjectID, pKeys);
        if (hr == S_OK)
        {
            hr = GetPropertyValuesForObject(wszObjectID, pKeys, pValues);
        }
    }

    // S_OK or S_FALSE can be returned from GetPropertyValuesForObject( ).
    // S_FALSE means that 1 or more property values could not be retrieved successfully.
    // The value for the specified property key should be set to the error HRESULT of
    // the reason why the property could not be read.
    // (i.e. an error of HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED) if a property value was
    //  requested and is not supported by the specified object.)
    if (SUCCEEDED(hr))
    {
        // Set the WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_VALUES value in the results
        HRESULT hrTemp = S_OK;
        hrTemp = pResults->SetIPortableDeviceValuesValue(WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_VALUES, pValues);

        if(FAILED(hrTemp))
        {
            hr = hrTemp;
        }
    }

    // Free the memory.  CoTaskMemFree ignores NULLs so no need to check.
    CoTaskMemFree(wszObjectID);

    return hr;
}
*/

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_SET
 *  command.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose property values we want to return.
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_VALUES: an IPortableDeviceValues of values, identifying which
 *      specific property values we are requested to write.
 *
 *  The driver should:
 *  - Write all requested property values.  For each property, a write result should be returned in the
 *      write result property store.
 *  - If any property write failed, the corresponding write result value should be
 *      set to type VT_ERROR with the 'scode' member holding the HRESULT reason for the failure.
 *  - S_OK should be returned if all properties were written successfully.
 *  - S_FALSE should be returned if any property write failed.
 *  - Any error return indicates that the driver did not write any results, and the caller will
 *      not attempt to unpack any property write results.
 */
HRESULT WpdObjectProperties::OnSetPropertyValues(
        __in LPWSTR  wszObjectID,
        IPortableDeviceValues* pValues,
        IPortableDeviceValues** ppResults
        )
{
    HRESULT         hr          = S_OK;
    DWORD           cValues     = 0;
    BOOL            fHasError   = FALSE;
    SensorType      sType       = SensorNone;

    // CoCreate a collection to store the property set operation results.
    hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**) ppResults);

    // Get the sensor type 
    if( hr == S_OK )
    {        
        hr = FindSensorTypeFromObjectID(wszObjectID, &sType);
    }

    // Set the property values on the specified object
    if (hr == S_OK)
    {
        hr = pValues->GetCount(&cValues);

        if (hr == S_OK)
        {
            for (DWORD dwIndex = 0; dwIndex < cValues; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                PROPVARIANT var;

                PropVariantInit( &var );

                hr = pValues->GetAt(dwIndex, &Key, &var);

                if(hr == S_OK)
                {
                    // Check if this is a supported settable property
                    if(TRUE == IsSettablePropertyKey(Key))
                    {
                        // Set the value
                        hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(sType))->SetProperty(Key, &var);
                    }
                    else
                    {
                        hr = (*ppResults)->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_FOUND));
                        fHasError = TRUE;
                    }
                }

                PropVariantClear(&var);
            
                if( FAILED( hr ) )
                {
                    break;
                }
            
            }
        }

        // Since we have set failures for the property set operations we must let the application
        // know by returning S_FALSE. This will instruct the application to look at the
        // property set operation results for failure values.
        if( (hr == S_OK) && (TRUE == fHasError) )
        {
            hr = S_FALSE;
        }
    }

    return hr;
}

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_GET_ATTRIBUTES
 *  command.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose property attributes we want to return.
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_KEYS: a collection of property keys containing a single value,
 *      which is the key identifying the specific property attributes we are requested to return.
 *
 *  The driver should:
 *  - Return the requested property attributes.  If any property attributes failed to be retrieved,
 *    the corresponding value should be set to type VT_ERROR with the 'scode' member holding the
 *    HRESULT reason for the failure.
 *  - S_OK should be returned if all property attributes were read successfully.
 *  - S_FALSE should be returned if any property attribute failed.
 *  - Any error return indicates that the driver did not fill in any results, and the caller will
 *      not attempt to unpack any property values.
 */
HRESULT WpdObjectProperties::OnGetPropertyAttributes(
    IPortableDeviceValues*  pParams,
    IPortableDeviceValues*  pResults)
{
    HRESULT     hr          = S_OK;
    LPWSTR      wszObjectID = NULL;
    PROPERTYKEY Key         = WPD_PROPERTY_NULL;
    CComPtr<IPortableDeviceValues>  pAttributes;

    // First get ALL parameters for this command.  If we cannot get ALL parameters
    // then E_INVALIDARG should be returned and no further processing should occur.

    // Get the object identifier whose property attributes have been requested
    if (hr == S_OK)
    {
        hr = pParams->GetStringValue(WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID, &wszObjectID);
    }

    // Get the list of property keys whose attributes are being requested
    if (hr == S_OK)
    {
        hr = pParams->GetKeyValue(WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_KEYS, &Key);
    }

    // CoCreate a collection to store the property attributes.
    if (hr == S_OK)
    {
        hr = CoCreateInstance(CLSID_PortableDeviceValues,
                              NULL,
                              CLSCTX_INPROC_SERVER,
                              IID_IPortableDeviceValues,
                              (VOID**) &pAttributes);
    }

    // Get the attributes for the specified properties on the specified object and add them
    // to the collection.
    if (hr == S_OK)
    {
        hr = GetPropertyAttributesForObject(wszObjectID, Key, pAttributes);
    }

    if (SUCCEEDED(hr))
    {
        // Set the WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_ATTRIBUTES value in the results
        HRESULT hrTemp = S_OK;
        hrTemp = pResults->SetIPortableDeviceValuesValue(WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_ATTRIBUTES, pAttributes);

        if(FAILED(hrTemp))
        {
            hr = hrTemp;
        }
    }

    // Free the memory.  CoTaskMemFree ignores NULLs so no need to check.
    CoTaskMemFree(wszObjectID);

    return hr;
}

/**
 *  This method is called when we receive a WPD_COMMAND_OBJECT_PROPERTIES_DELETE
 *  command.
 *
 *  The parameters sent to us are:
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_OBJECT_ID: identifies the object whose properties should be deleted.
 *  - WPD_PROPERTY_OBJECT_PROPERTIES_PROPERTY_KEYS: a collection of property keys indicating which
 *      properties to delete.
 *
 *  The driver should:
 *  - Delete the specified properties from the object.
 *  - S_OK should be returned if all specified properties were successfully deleted.
 *  - E_ACCESSDENIED should be returned if the client attempts to delete a property which is not deletable (i.e.
 *    WPD_PROPERTY_ATTRIBUTE_CAN_DELETE is FALSE for that property.)
 */
HRESULT WpdObjectProperties::OnDeleteProperties(
    IPortableDeviceValues*  pParams,
    IPortableDeviceValues*  pResults)
{
    HRESULT hr = E_ACCESSDENIED;

    UNREFERENCED_PARAMETER(pParams);
    UNREFERENCED_PARAMETER(pResults);

    // This driver has no properties which can be deleted.

    return hr;
}

/**
 *  This method is called to populate property values for the object specified.
 *
 *  The parameters sent to us are:
 *  wszObjectID - the object whose properties are being requested.
 *  pKeys - the list of property keys of the properties to request from the object
 *  pValues - an IPortableDeviceValues which will contain the property values retreived from the object
 *
 *  The driver should:
 *  Read the specified properties for the specified object and populate pValues with the
 *  results.
 */
HRESULT WpdObjectProperties::GetPropertyValuesForObject(
    LPCWSTR                        wszObjectID,
    IPortableDeviceKeyCollection*  pKeys,
    IPortableDeviceValues*         pValues)
{
    HRESULT     hr          = S_OK;
    CAtlStringW strObjectID = wszObjectID;
    DWORD       cKeys       = 0;
    BOOL        fError      = FALSE;

    if ((wszObjectID == NULL) ||
        (pKeys       == NULL) ||
        (pValues     == NULL))
    {
        hr = E_INVALIDARG;
        return hr;
    }

    hr = pKeys->GetCount(&cKeys);

    if (hr == S_OK)
    {
        // Get values for the DEVICE object
        if (strObjectID.CompareNoCase(WPD_DEVICE_OBJECT_ID) == 0)
        {
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK && !IsEqualPropertyKey(Key, WPD_PROPERTY_NULL))
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));
                  
                    // Set DEVICE object properties
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_FIRMWARE_VERSION))
                    {
                        hr = pValues->SetStringValue(WPD_DEVICE_FIRMWARE_VERSION, DEVICE_FIRMWARE_VERSION_VALUE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_POWER_LEVEL))
                    {
                        hr = pValues->SetUnsignedIntegerValue(WPD_DEVICE_POWER_LEVEL, DEVICE_POWER_LEVEL_VALUE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_POWER_SOURCE))
                    {
                        hr = pValues->SetUnsignedIntegerValue(WPD_DEVICE_POWER_SOURCE, WPD_POWER_SOURCE_EXTERNAL);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_PROTOCOL))
                    {
                        hr = pValues->SetStringValue(WPD_DEVICE_PROTOCOL, DEVICE_PROTOCOL_VALUE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_MODEL))
                    {
                        hr = pValues->SetStringValue(WPD_DEVICE_MODEL, g_wszDeviceName);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_SERIAL_NUMBER))
                    {
                        hr = pValues->SetStringValue(WPD_DEVICE_SERIAL_NUMBER, DEVICE_SERIAL_NUMBER_VALUE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_SUPPORTS_NON_CONSUMABLE))
                    {
                        hr = pValues->SetBoolValue(WPD_DEVICE_SUPPORTS_NON_CONSUMABLE, DEVICE_SUPPORTS_NONCONSUMABLE_VALUE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_MANUFACTURER))
                    {
                        hr = pValues->SetStringValue(WPD_DEVICE_MANUFACTURER, DEVICE_MANUFACTURER_VALUE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_FRIENDLY_NAME))
                    {
                        hr = pValues->SetStringValue(WPD_DEVICE_FRIENDLY_NAME, DEVICE_FRIENDLY_NAME_VALUE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_DEVICE_TYPE))
                    {
                        hr = pValues->SetUnsignedIntegerValue(WPD_DEVICE_TYPE, WPD_DEVICE_TYPE_GENERIC);
                    }
                    else
                    // Set general properties for DEVICE
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_ID, WPD_DEVICE_OBJECT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_NAME))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_NAME, WPD_DEVICE_OBJECT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PERSISTENT_UNIQUE_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PERSISTENT_UNIQUE_ID, WPD_DEVICE_OBJECT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PARENT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PARENT_ID, L"");
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_FORMAT))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_FORMAT, WPD_OBJECT_FORMAT_UNSPECIFIED);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CONTENT_TYPE))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_CONTENT_TYPE, WPD_CONTENT_TYPE_FUNCTIONAL_OBJECT);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CAN_DELETE))
                    {
                        hr = pValues->SetBoolValue(WPD_OBJECT_CAN_DELETE, FALSE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_FUNCTIONAL_OBJECT_CATEGORY))
                    {
                        hr = pValues->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, WPD_FUNCTIONAL_CATEGORY_DEVICE);
                    }
                }
            }
		}
        else if (strObjectID.CompareNoCase(SENSOR_ACCELEROMETER_ID) == 0)
		{
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK && !IsEqualPropertyKey(Key, WPD_PROPERTY_NULL))
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));

                    // Set general properties for sensor
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_ID, SENSOR_ACCELEROMETER_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_NAME))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_NAME, SENSOR_ACCELEROMETER_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PERSISTENT_UNIQUE_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PERSISTENT_UNIQUE_ID, SENSOR_ACCELEROMETER_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PARENT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PARENT_ID, WPD_DEVICE_OBJECT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_FORMAT))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_FORMAT, WPD_OBJECT_FORMAT_UNSPECIFIED);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CONTENT_TYPE))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_CONTENT_TYPE, WPD_CONTENT_TYPE_FUNCTIONAL_OBJECT);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CAN_DELETE))
                    {
                        hr = pValues->SetBoolValue(WPD_OBJECT_CAN_DELETE, FALSE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_FUNCTIONAL_OBJECT_CATEGORY))
                    {
                        hr = pValues->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_MOTION);
                    }
                    else
                    {   
                        // Get accelerometer sensor properties
                        PROPVARIANT value;
                        PropVariantInit(&value);

                        HRESULT hrTemp = g_pSensorList.GetAt(g_pSensorList.FindIndex(Accelerometer))->GetProperty(Key, &value);
                        if (SUCCEEDED(hrTemp))
                        {
                            pValues->SetValue(Key, &value);
                        }
                        else
                        {
                            // Failed to find the requested property, convey the hr back to the caller
                            pValues->SetErrorValue(Key, hrTemp);
                            fError = TRUE;
                        }

                        PropVariantClear(&value);

                    }
                }
            }
		}
        else if (strObjectID.CompareNoCase(SENSOR_AMBIENT_LIGHT_ID) == 0)
		{
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK && !IsEqualPropertyKey(Key, WPD_PROPERTY_NULL))
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));

                     // Set general properties for sensor
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_ID, SENSOR_AMBIENT_LIGHT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_NAME))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_NAME, SENSOR_AMBIENT_LIGHT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PERSISTENT_UNIQUE_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PERSISTENT_UNIQUE_ID, SENSOR_AMBIENT_LIGHT_ID);
					}
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PARENT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PARENT_ID, WPD_DEVICE_OBJECT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_FORMAT))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_FORMAT, WPD_OBJECT_FORMAT_UNSPECIFIED);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CONTENT_TYPE))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_CONTENT_TYPE, WPD_CONTENT_TYPE_FUNCTIONAL_OBJECT);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CAN_DELETE))
                    {
                        hr = pValues->SetBoolValue(WPD_OBJECT_CAN_DELETE, FALSE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_FUNCTIONAL_OBJECT_CATEGORY))
                    {
                        hr = pValues->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_LIGHT);
                    }
                    else
                    {
                        // Get ambient light sensor properties
                        PROPVARIANT value;
                        PropVariantInit(&value);

                        HRESULT hrTemp = g_pSensorList.GetAt(g_pSensorList.FindIndex(AmbientLight))->GetProperty(Key, &value);
                        if (SUCCEEDED(hrTemp))
                        {
                            pValues->SetValue(Key, &value);
                        }
                        else
                        {
                            //
                            // Failed to find the requested property, convey the hr back to the caller
                            //
                            pValues->SetErrorValue(Key, hrTemp);
                            fError = TRUE;
                        }

                        PropVariantClear(&value);
                    }
                }
            }
		}
        else if (strObjectID.CompareNoCase(SENSOR_SWITCH_ARRAY_LEFT_ID) == 0)
		{
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK && !IsEqualPropertyKey(Key, WPD_PROPERTY_NULL))
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));

                    // Set general properties for sensor
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_ID, SENSOR_SWITCH_ARRAY_LEFT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_NAME))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_NAME, SENSOR_SWITCH_ARRAY_LEFT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PERSISTENT_UNIQUE_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PERSISTENT_UNIQUE_ID, SENSOR_SWITCH_ARRAY_LEFT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PARENT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PARENT_ID, WPD_DEVICE_OBJECT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_FORMAT))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_FORMAT, WPD_OBJECT_FORMAT_UNSPECIFIED);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CONTENT_TYPE))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_CONTENT_TYPE, WPD_CONTENT_TYPE_FUNCTIONAL_OBJECT);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CAN_DELETE))
                    {
                        hr = pValues->SetBoolValue(WPD_OBJECT_CAN_DELETE, FALSE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_FUNCTIONAL_OBJECT_CATEGORY))
                    {
                        hr = pValues->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_MECHANICAL);
                    }
                    else
                    {
                        // Get switch array left sensor properties
                        PROPVARIANT value;
                        PropVariantInit(&value);

                        HRESULT hrTemp = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayLeft))->GetProperty(Key, &value);
                        if (SUCCEEDED(hrTemp))
                        {
                            pValues->SetValue(Key, &value);
                        }
                        else
                        {
                            //
                            // Failed to find the requested property, convey the hr back to the caller
                            //
                            pValues->SetErrorValue(Key, hrTemp);
                            fError = TRUE;
                        }

                        PropVariantClear(&value);
                    }
                }
            }
		}
        else if (strObjectID.CompareNoCase(SENSOR_SWITCH_ARRAY_RIGHT_ID) == 0)
		{
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK && !IsEqualPropertyKey(Key, WPD_PROPERTY_NULL))
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));

                    // Set general properties for sensor
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_ID, SENSOR_SWITCH_ARRAY_RIGHT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_NAME))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_NAME, SENSOR_SWITCH_ARRAY_RIGHT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PERSISTENT_UNIQUE_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PERSISTENT_UNIQUE_ID, SENSOR_SWITCH_ARRAY_RIGHT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_PARENT_ID))
                    {
                        hr = pValues->SetStringValue(WPD_OBJECT_PARENT_ID, WPD_DEVICE_OBJECT_ID);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_FORMAT))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_FORMAT, WPD_OBJECT_FORMAT_UNSPECIFIED);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CONTENT_TYPE))
                    {
                        hr = pValues->SetGuidValue(WPD_OBJECT_CONTENT_TYPE, WPD_CONTENT_TYPE_FUNCTIONAL_OBJECT);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_OBJECT_CAN_DELETE))
                    {
                        hr = pValues->SetBoolValue(WPD_OBJECT_CAN_DELETE, FALSE);
                    }
                    else
                    if (IsEqualPropertyKey(Key, WPD_FUNCTIONAL_OBJECT_CATEGORY))
                    {
                        hr = pValues->SetGuidValue(WPD_FUNCTIONAL_OBJECT_CATEGORY, SENSOR_CATEGORY_MECHANICAL);
                    }
                    else
                    {
                        // Get switch array right sensor properties
                        PROPVARIANT value;
                        PropVariantInit(&value);

                        HRESULT hrTemp = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayRight))->GetProperty(Key, &value);
                        if (SUCCEEDED(hrTemp))
                        {
                            pValues->SetValue(Key, &value);
                        }
                        else
                        {
                            //
                            // Failed to find the requested property, convey the hr back to the caller
                            //
                            pValues->SetErrorValue(Key, hrTemp);
                            fError = TRUE;
                        }

                        PropVariantClear(&value);
                    }
                }
            }
        }
    }

    return (FALSE == fError) ? hr : S_FALSE;
}

/**
 *  This method is called to populate data values for the object specified.
 *
 *  The parameters sent to us are:
 *  wszObjectID - the object whose properties are being requested.
 *  pKeys - the list of property keys of the properties to request from the object
 *  pValues - an IPortableDeviceValues which will contain the property values retreived from the object
 *
 *  The driver should:
 *  Read the specified properties for the specified object and populate pValues with the
 *  results.
 */
HRESULT WpdObjectProperties::GetDataValuesForObject(
    LPCWSTR                        wszObjectID,
    IPortableDeviceKeyCollection*  pKeys,
    IPortableDeviceValues*         pValues)
{
    HRESULT     hr          = S_OK;
    CAtlStringW strObjectID = wszObjectID;
    DWORD       cKeys       = 0;

    if ((wszObjectID == NULL) ||
        (pKeys       == NULL) ||
        (pValues     == NULL))
    {
        hr = E_INVALIDARG;
        return hr;
    }

    hr = pKeys->GetCount(&cKeys);

    if (hr == S_OK)
    {
		if (strObjectID.CompareNoCase(SENSOR_ACCELEROMETER_ID) == 0)
		{
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK)
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));
                    
                    PROPVARIANT var;
                    PropVariantInit( &var );

                    hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(Accelerometer))->GetDataField(Key, &var);
                        
                    if (SUCCEEDED(hr))
                    {
                        hr = pValues->SetValue(Key, &var);
                    }

                    // Free any allocated resources
                    PropVariantClear( &var );

                }
            }
		}
		else if (strObjectID.CompareNoCase(SENSOR_AMBIENT_LIGHT_ID) == 0)
		{
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK)
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));
                    
                    PROPVARIANT var;
                    PropVariantInit( &var );

                    hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(AmbientLight))->GetDataField(Key, &var);
                        
                    if (SUCCEEDED(hr))
                    {
                        hr = pValues->SetValue(Key, &var);
                    }

                    // Free any allocated resources
                    PropVariantClear( &var );

                }
            }
		}
		else if (strObjectID.CompareNoCase(SENSOR_SWITCH_ARRAY_LEFT_ID) == 0)
		{
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK)
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));
                    
                    PROPVARIANT var;
                    PropVariantInit( &var );

                    hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayLeft))->GetDataField(Key, &var);
                        
                    if (SUCCEEDED(hr))
                    {
                        hr = pValues->SetValue(Key, &var);
                    }

                    // Free any allocated resources
                    PropVariantClear( &var );
                }
            }
		}
        else if (strObjectID.CompareNoCase(SENSOR_SWITCH_ARRAY_RIGHT_ID) == 0)
        {
            for (DWORD dwIndex = 0; dwIndex < cKeys; dwIndex++)
            {
                PROPERTYKEY Key = WPD_PROPERTY_NULL;
                hr = pKeys->GetAt(dwIndex, &Key);

                if (hr == S_OK)
                {
                    // Preset the property value to 'error not supported'.  The actual value
                    // will replace this value, if read from the device.
                    pValues->SetErrorValue(Key, HRESULT_FROM_WIN32(ERROR_NOT_SUPPORTED));
                    
                    PROPVARIANT var;
                    PropVariantInit( &var );

                    hr = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayRight))->GetDataField(Key, &var);
                        
                    if (SUCCEEDED(hr))
                    {
                        hr = pValues->SetValue(Key, &var);
                    }

                    // Free any allocated resources
                    PropVariantClear( &var );
                }
            }
		}
    }

    return hr;
}

/**
 *  This method is called to populate property attributes for the object and property specified.
 *
 *  The parameters sent to us are:
 *  wszObjectID - the object whose property attributes are being requested.
 *  Key - the property whose attributes are being requested
 *  pAttributes - an IPortableDeviceValues which will contain the resulting property attributes
 *
 *  The driver should:
 *  Read the property attributes for the specified property on the specified object and
 *  populate pAttributes with the results.
 */
HRESULT WpdObjectProperties::GetPropertyAttributesForObject(
    LPCWSTR                wszObjectID,
    const PROPERTYKEY&     Key,
    IPortableDeviceValues* pAttributes)
{
    HRESULT hr = S_OK;

    if ((wszObjectID == NULL) ||
        (pAttributes == NULL))
    {
        hr = E_INVALIDARG;
        return hr;
    }

    UNREFERENCED_PARAMETER(wszObjectID);
    UNREFERENCED_PARAMETER(Key);

    //
    // Since ALL of our properties have the same attributes, we are ignoring the
    // passed in wszObjectID and Key parameters.  These parameters allow you to
    // customize attributes for properties on specific objects. (i.e. WPD_OBJECT_ORIGINAL_FILE_NAME
    // may be READ/WRITE on some objects and READONLY on others. )
    //

    if (hr == S_OK)
    {
        hr = pAttributes->SetBoolValue(WPD_PROPERTY_ATTRIBUTE_CAN_DELETE, FALSE);
    }

    if (hr == S_OK)
    {
        hr = pAttributes->SetBoolValue(WPD_PROPERTY_ATTRIBUTE_CAN_READ, TRUE);
    }

    if (hr == S_OK)
    {
        hr = pAttributes->SetBoolValue(WPD_PROPERTY_ATTRIBUTE_CAN_WRITE, FALSE);
    }

    if (hr == S_OK)
    {
        hr = pAttributes->SetBoolValue(WPD_PROPERTY_ATTRIBUTE_FAST_PROPERTY, TRUE);
    }

    if (hr == S_OK)
    {
        hr = pAttributes->SetUnsignedIntegerValue(WPD_PROPERTY_ATTRIBUTE_FORM, WPD_PROPERTY_ATTRIBUTE_FORM_UNSPECIFIED);
    }

    return hr;
}

/**
 *  This method is called by DDI when a client subscribes/unsubscribes to events
 *
 *  The parameters sent to us are:
 *  bClientSubscribed - A Boolean value that denotes the subscription status
 *
 *  Action:
 *  Set the Client Status on the CLightSensorDevice object
 */
HRESULT WpdObjectProperties::OnClientSubscribeToEvents(__in LPWSTR pwszObjectID)
{
    HRESULT hr = S_OK;

    // Set the Client Status : denotes the status of event subscribed clients
    SensorType sType;
    hr = FindSensorTypeFromObjectID(pwszObjectID, &sType);

    if(SUCCEEDED(hr))
    {
        g_pSensorList.GetAt(g_pSensorList.FindIndex(sType))->Subscribe();
    }

    return hr;
}


HRESULT WpdObjectProperties::OnClientUnsubscribeToEvents(__in LPWSTR pwszObjectID)
{
    HRESULT hr = S_OK;

    // Set the Client Status : denotes the status of event subscribed clients
    SensorType sType;
    hr = FindSensorTypeFromObjectID(pwszObjectID, &sType);

    if(SUCCEEDED(hr))
    {
        g_pSensorList.GetAt(g_pSensorList.FindIndex(sType))->Unsubscribe();
    }

    return hr;
}

HRESULT WpdObjectProperties::ConvertByteToFloat(
	__in DWORD Id,
	__in BYTE  ByteData,
	__out FLOAT *pFloatData)
{
    HRESULT hr = (NULL != pFloatData) ? S_OK : E_UNEXPECTED;
	DWORD   iValue;

    if(SUCCEEDED(hr))
	{
		if (Id == Accelerometer)
		{
			iValue = ByteData;
			*pFloatData = (iValue * 4.0F / 255) - 2.0F;
		}
		else if (Id == AmbientLight)
		{
			iValue = ByteData;
#if 0
			*pFloatData = iValue * 4.0F / 255;
#endif
#if 0
			*pFloatData = iValue * 316.22F / 255;
			*pFloatData *= *pFloatData;
#endif
#if 1
			*pFloatData = iValue * 5.0F / 255;
			*pFloatData = powf(10.0F, *pFloatData);
#endif
		}
		else
		{
			*pFloatData = 0.0F;
			hr = E_INVALIDARG;
		}
	}

    return hr;
}

#if 0
//This function parses the content in the buffer and updates cached data vlaues
HRESULT WpdObjectProperties::ProcessAsynchRead( BYTE* buffer, ULONG length )
{
    UNREFERENCED_PARAMETER(length);

    HRESULT             hr = S_OK;
    CBadgeSensorDevice* pSensor = NULL;

    
    DWORD               bufOffset  = 1;
    IN_STRUCT*          recv_report = NULL;
    
    // Add 1 byte offset because USB adds a leading byte
    recv_report = (IN_STRUCT*) ( buffer + bufOffset );

    // Check the length is non-zero
    //if(length > 0)
    {
        // Device ID = 1 set in firmware
        //if(recv_report->device_id == 1)
        if(1)
        {
            // Subtract 1 because we ignore Base Device
            DWORD sensorId = (recv_report->sensor_id >> 3) - 1;
            DWORD sensorCount = recv_report->sensor_id & 0x7;

            switch(sensorId)
            {
            case Accelerometer:
                
                pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(Accelerometer));
                
                // Compare the GUIDs
                //if(IsEqualGUID(recv_report->report_id, SENSOR_DATA_TYPE_MOTION_GUID) &&
                //    (recv_report->prop_variant == 1))
				if (sensorCount == 3)
                {
                    hr = pSensor->SetTimeStamp();

                    if(SUCCEEDED(hr))
                    {
                        PROPVARIANT value;
                        PropVariantInit( &value );
                        value.vt = VT_R4;

                        //value.fltVal = recv_report->float_data[0];
                        hr = ConvertByteToFloat(sensorId, recv_report->byte_data[0], &value.fltVal);

						if(SUCCEEDED(hr))
                        {
	                        hr = pSensor->SetDataField(SENSOR_DATA_TYPE_ACCELERATION_X_G, &value);
						}

						if(SUCCEEDED(hr))
                        {
	                        hr = ConvertByteToFloat(sensorId, recv_report->byte_data[1], &value.fltVal);
						}
						if(SUCCEEDED(hr))
                        {
                            //value.fltVal = recv_report->float_data[1];
                            hr = pSensor->SetDataField(SENSOR_DATA_TYPE_ACCELERATION_Y_G, &value);
                        }

						if(SUCCEEDED(hr))
                        {
	                        hr = ConvertByteToFloat(sensorId, recv_report->byte_data[2], &value.fltVal);
						}
                        if(SUCCEEDED(hr))
                        {
                            //value.fltVal = recv_report->float_data[2];
                            hr = pSensor->SetDataField(SENSOR_DATA_TYPE_ACCELERATION_Z_G, &value);
                        }

                        PropVariantClear( &value );
                    }

                    if( SUCCEEDED(hr))
                    {
                        pSensor->RaiseDataEvent();
                    }

                }

                break;

            case AmbientLight:
                
                pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(AmbientLight));
                
                // Compare the GUIDs
                //if(IsEqualGUID(recv_report->report_id, SENSOR_DATA_TYPE_LIGHT_GUID) &&
                //    (recv_report->prop_variant == 2))
				if (sensorCount == 1)
                {
                    
                    hr = pSensor->SetTimeStamp();

                    if(SUCCEEDED(hr))
                    {
                        PROPVARIANT value;
                        PropVariantInit( &value );
                        value.vt = VT_R4;

                        //value.fltVal = recv_report->float_data[0];
                        hr = ConvertByteToFloat(sensorId, recv_report->byte_data[0], &value.fltVal);
                        if( SUCCEEDED(hr))
                        {
	                        hr = pSensor->SetDataField(SENSOR_DATA_TYPE_LIGHT_LEVEL_LUX, &value);
						}
                        PropVariantClear( &value );

                        if( SUCCEEDED(hr))
                        {
                            pSensor->RaiseDataEvent();
                        }
                    }
                    
                }

                break;

            case SwitchArrayLeft:

                pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayLeft));
                
                // Compare the GUIDs
                //if(IsEqualGUID(recv_report->report_id, SENSOR_DATA_TYPE_GUID_MECHANICAL_GUID) &&
                //    (recv_report->prop_variant == 10))
				if (sensorCount == 1)
                {
                    hr = pSensor->SetTimeStamp();

                    if(SUCCEEDED(hr))
                    {
                        PROPVARIANT value;
                        PropVariantInit( &value );
                        value.vt = VT_UI4;

                        //value.ulVal = (ULONG) recv_report->dword_data[0];
                        value.ulVal = (ULONG) (recv_report->byte_data[0] & 0xF);
                        hr = pSensor->SetDataField(SENSOR_DATA_TYPE_BOOLEAN_SWITCH_ARRAY_STATE, &value);

                        PropVariantClear( &value );
                    }

                    if( SUCCEEDED(hr))
                    {
                        pSensor->RaiseDataEvent();
                    }
                }

            //    break;
            //
            //case SwitchArrayRight:

                pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayRight));
                
                // Compare the GUIDs
                //if(IsEqualGUID(recv_report->report_id, SENSOR_DATA_TYPE_GUID_MECHANICAL_GUID) &&
                //    (recv_report->prop_variant == 10))
				if (sensorCount == 1)
                {
                    hr = pSensor->SetTimeStamp();

                    if(SUCCEEDED(hr))
                    {
                        PROPVARIANT value;
                        PropVariantInit( &value );
                        value.vt = VT_UI4;

                        //value.ulVal = (ULONG) recv_report->dword_data[0];
                        value.ulVal = (ULONG) (recv_report->byte_data[0] >> 4);
                        hr = pSensor->SetDataField(SENSOR_DATA_TYPE_BOOLEAN_SWITCH_ARRAY_STATE, &value);

                        PropVariantClear( &value );
                    }

                    if( SUCCEEDED(hr))
                    {
                        pSensor->RaiseDataEvent();
                    }
                }

                break;

            default:

                break;

            }
        }
    }

    return hr;
}
#endif

HRESULT WpdObjectProperties::ProcessAsynchRead( BYTE* buffer, ULONG length )
{
    UNREFERENCED_PARAMETER(length);

    HRESULT             hr = S_OK;
    CBadgeSensorDevice* pSensor = NULL;
    DWORD               bufOffset  = 1;

    // Add 1 byte offset because USB adds a leading byte
    IN_STRUCT*          recv_report = (IN_STRUCT*) ( buffer + bufOffset );

    // Subtract 1 because we ignore Base Device
    DWORD sensorId = (recv_report->sensor_id >> 3) - 1;
    DWORD sensorCount = recv_report->sensor_id & 0x7;

    switch(sensorId)
    {
    case Accelerometer:

        pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(Accelerometer));

        if (sensorCount == 3)
        {
            hr = pSensor->SetTimeStamp();

            if(SUCCEEDED(hr))
            {
                PROPVARIANT value;
                PropVariantInit( &value );
                value.vt = VT_R4;

                hr = ConvertByteToFloat(sensorId, recv_report->byte_data[0], &value.fltVal);

                if(SUCCEEDED(hr))
                {
                    hr = pSensor->SetDataField(SENSOR_DATA_TYPE_ACCELERATION_X_G, &value);
                }
                if(SUCCEEDED(hr))
                {
                    hr = ConvertByteToFloat(sensorId, recv_report->byte_data[1], &value.fltVal);
                }
                if(SUCCEEDED(hr))
                {
                    hr = pSensor->SetDataField(SENSOR_DATA_TYPE_ACCELERATION_Y_G, &value);
                }
                if(SUCCEEDED(hr))
                {
                    hr = ConvertByteToFloat(sensorId, recv_report->byte_data[2], &value.fltVal);
                }
                if(SUCCEEDED(hr))
                {
                    hr = pSensor->SetDataField(SENSOR_DATA_TYPE_ACCELERATION_Z_G, &value);
                }
                PropVariantClear( &value );
            }
            if( SUCCEEDED(hr))
            {
                pSensor->RaiseDataEvent();
            }
        }
        break;

    case AmbientLight:

        pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(AmbientLight));

        if (sensorCount == 1)
        {
            hr = pSensor->SetTimeStamp();

            if(SUCCEEDED(hr))
            {
                PROPVARIANT value;
                PropVariantInit( &value );
                value.vt = VT_R4;

                hr = ConvertByteToFloat(sensorId, recv_report->byte_data[0], &value.fltVal);
                if( SUCCEEDED(hr))
                {
                    hr = pSensor->SetDataField(SENSOR_DATA_TYPE_LIGHT_LEVEL_LUX, &value);
                }
                PropVariantClear( &value );

                if( SUCCEEDED(hr))
                {
                    pSensor->RaiseDataEvent();
                }
            }

        }

        break;

    case SwitchArrayLeft:

        pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayLeft));

        if (sensorCount == 1)
        {
            hr = pSensor->SetTimeStamp();

            if(SUCCEEDED(hr))
            {
                PROPVARIANT value;
                PropVariantInit( &value );
                value.vt = VT_UI4;

                value.ulVal = (ULONG) (recv_report->byte_data[0] & 0xF);
                hr = pSensor->SetDataField(SENSOR_DATA_TYPE_BOOLEAN_SWITCH_ARRAY_STATE, &value);

                PropVariantClear( &value );
            }

            if( SUCCEEDED(hr))
            {
                pSensor->RaiseDataEvent();
            }
        }

        pSensor = g_pSensorList.GetAt(g_pSensorList.FindIndex(SwitchArrayRight));

        if (sensorCount == 1)
        {
            hr = pSensor->SetTimeStamp();

            if(SUCCEEDED(hr))
            {
                PROPVARIANT value;
                PropVariantInit( &value );
                value.vt = VT_UI4;

                value.ulVal = (ULONG) (recv_report->byte_data[0] >> 4);
                hr = pSensor->SetDataField(SENSOR_DATA_TYPE_BOOLEAN_SWITCH_ARRAY_STATE, &value);

                PropVariantClear( &value );
            }

            if( SUCCEEDED(hr))
            {
                pSensor->RaiseDataEvent();
            }
        }

        break;

    default:

        break;

    }

    return hr;
}
